/**
 * Singleton instance of the
 * Icon Manager.  Instantiated
 * at bottom of file.
 */
var gIconManager;

function IconManager() {
  if (! IconManager._hasRequiredComponents()) {
      return;
  }

  // whether sidebar is active
  this.sidebarActive = false;

  // initialize annotations service
  this.annotService = Components.classes[IconManager.ANNOTATIONS_SERVICE_CONTRACTID].getService().QueryInterface(Components.interfaces.nsIAnnotationsService);
  
  // get the annozilla AnnotationsApplication object
  this.app = this.annotService.registerApplication("annozilla");

  // whether we're inserting icons
  this.insertIcons = false;

  // say that we are not observing annotations -- this must be done before loading prefs
  this.observing = false;

  // prefs module
  this.prefs = new AnnozillaPrefs(this);
  
  // bundle
  this.bundle = srGetStrBundle("chrome://annozilla/locale/annozilla.properties");

  var that = this;

  // add a listener for when the window closes
  window.addEventListener("unload",
			  function(e) { that.unload(); },
			    false);

  // initialize annotations
  this.annotations = {};
  this.currentURI = null;
  this.currentURIState = null;
  this.userLoadInProgress = false;
}

IconManager.PAGE_LOADING = 0;
IconManager.PAGE_COMPLETED = 1;
IconManager.PAGE_FAILED = 2;

IconManager.ANNOTATIONS_SERVICE_CONTRACTID = "@mozilla.org/annotations_service;1";
IconManager.XPOINTER_SERVICE_CONTRACTID = "@mozilla.org/xpointer-service;1";

IconManager.ANNOT_NS = "http://www.w3.org/2000/10/annotation-ns#";
IconManager.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
IconManager.DC10_NS    = "http://purl.org/dc/elements/1.0/";
IconManager.DC11_NS    = "http://purl.org/dc/elements/1.1/";
IconManager.THREAD_NS= "http://www.w3.org/2001/03/thread#";
IconManager.XHTML_NS = "http://www.w3.org/1999/xhtml";
IconManager.TYPE_NS = "http://www.w3.org/2000/10/annotationType#";
IconManager.XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

/*******************
 * PUBLIC FUNCTIONS
 *******************/

IconManager.prototype.QueryInterface = function(iid) {  
  if (! (iid.equals(Components.interfaces.nsIAnnotationsObserver) ||
	 iid.equals(Components.interfaces.nsISupports) ) ) {
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
  return this;
}

IconManager.prototype.unload = function() {
    if (this.observing) {
	// remove observer
	this.app.removeObserver(this);
	this.observing = false;
    }

    // tell prefs we're shutting down
    this.prefs.destructor();
}
  
// called whenever the insertIcons pref is set
IconManager.prototype.insertIconsPrefChanging = function(insert) {
    this.insertIcons = insert;

    if (insert && (! this.observing) ) {
	// add our observer
	this.app.addObserver(this);
	this.observing = true;
    }
    if ((!insert) && this.observing) {
	// remove our observer
	this.app.removeObserver(this);
	this.observing = false;
    }
}  

IconManager.prototype.sbarOnLoad = function() {
  this.sidebarActive = true;

  // remove all annotations currently visible
   if (this._isInsertIconsEnabled()) {
	    // remove all annotations from the page
	    this._removeAnnotations();
   }
	
  this.annotations = {};
  this.currentURI = null;
}

IconManager.prototype.sbarOnUnload = function() {
  this.sidebarActive = false;
}

// may come from load or autoload
IconManager.prototype.sbarLoadedAnnotations = function(annotations, annotatedURI) {

  // add all annotations first
  for (uri in annotations) {
      var annotObj = annotations[uri];
      IconManager._addAnnotationToContainer(annotObj.annot, annotObj.range, null, this.annotations);
  }

  if (this._isInsertIconsEnabled()) {
      this._insertAnnotations(annotations);
  }
}

/**
 * Called when a user annotations load
 * begins.
 */
IconManager.prototype.sbarUserLoadBegin = function(newURI) {
   
    if (this._isInsertIconsEnabled()) {
	if (newURI == this.currentURI) {
	    // remove all annotations from the page
	    this._removeAnnotations();
	}
    }
	
    this.annotations = {};
    this.currentURI = newURI;
    this.userLoadInProgress = true;
}

/**
 * Called when a page begins loading in the 
 * browser if the application has 
 * autoLoad turned on.  
 *
 * Note: only called if autoLoad is enabled.
 *
 * @param url string of url loading
 */
IconManager.prototype.autoLoadOnBeginPageLoad = function(uri, win) {
    // if this applies to another window, ignore
    if (window.content != win) {
	return;
    }

    this.currentURI = uri;
    this.annotations = {};
    this.currentURIState = IconManager.PAGE_LOADING;
    this.userLoadInProgress = false;
}

/**
 * Called when a page successfully finishes loading.
 *
 * Note: only called if autoLoad is enabled.
 *
 * @param url string of url finished loading
 * @param annotations nsISimpleEnumeration of annotations for url
 */
IconManager.prototype.autoLoadOnSuccessPageLoad = function(url, win) {
    // if this applies to another window, ignore
    if (window.content != win) {
	return;
    }

    this.currentURIState = IconManager.PAGE_COMPLETED;
    
    // if the sidebar is inactive and we're inserting icons
    if (! this.sidebarActive && this._isInsertIconsEnabled()) {

	// resolve all outstanding annotations
	for(uri in this.annotations) {
	    var annotationObj = this.annotations[uri];
	    
	    if (! annotationObj.annot.data) {
		this._debugBadAnnotation(annotationObj.annot);
		continue;
	    }
	    
	    annotationObj.range = xptrResolveAnnotation(annotationObj.annot, window._content.document);
	}
	
	// insert the annotations
	this._insertAnnotations(this.annotations);	
    }
}

/**
 * Called when a page that was loading is
 * unable to load (or is cancelled by the user).
 *
 * Note: only called if autoLoad is enabled.
 *
 * @param url string of url aborted
 * @param message string error message
 */
IconManager.prototype.autoLoadOnFailPageLoad = function(url, win) {
    // if this applies to another window, ignore
    if (window.content != win) {
	return;
    }

    this.currentURIState = IconManager.PAGE_FAILED;
}


/**
 * Success callback function to return the results for an
 * automatic annotations load (autoLoad).
 * This function returns the annotations for a single server, designated by serverURI.  
 * It will be called once for each server that successfully responds.
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri that the annotations annotate
 * @param annotations nsISimpleEnumerator of nsIAnnotations for annotatedURI that
 *                    were on serverURI annotations server
 */
IconManager.prototype.autoLoadAnnotationsForServer = function(serverURI, win, annotatedURI, count, annotations) {
    // if this applies to another window, ignore
    if (window.content != win) {
	return;
    }

    // if there is a user load in progress, ignore these annotations
    if (this.userLoadInProgress) {
	return;
    }

    this._processAnnotationsWithAutoLoadInfo(this.currentURIState, annotations);
}

/**
 * Callback function to indicate failure for an automatic
 * annotations load (autoLoad).
 * This function will be called once for each annotations server that
 * fails in the autoloading process. 
 *
 * @param serverURI string uri of the annotations server that failed
 * @param annotatedURI string uri for which annotations were requested
 * @param message string error message
 */ 
IconManager.prototype.autoLoadFailureForServer = function(serverURI, win, annotatedURI, message) {}


/**
 * Success callback function to return the results for a loadAnnotations request
 *  - OR - an automatic autoLoad request.  This function returns the 
 * annotations for one server, designated by serverURI.  
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri that the annotations annotate
 * @param annotations nsISimpleEnumerator of nsIAnnotations for annotatedURI that
 *                    were on serverURI annotations server
 */
IconManager.prototype.loadAnnotationsOnSuccess = function(serverURI, annotatedURI, annotationsLength, annotations) {

    // if the sidebar is inactive
    if (! this.sidebarActive ) {

	// if we're autoLoading, ignore these annotations
	if (! this.userLoadInProgress) {
	    return;
	}
	
	// if we've moved on to a new page, ignore these annotations
	if (annotatedURI != this.currentURI) {
	    return;
	}
	 
	// if autoLoad is enabled, we can benefit from its alerts about when the page is
	// fully loaded
	if (this._isAutoLoadEnabled()) {
	    this._processAnnotationsWithAutoLoadInfo(this.currentURIState, annotations);
	}
	// otherwise, we have to wing it. for now we'll assume that the page
	// is loaded, since there's no way that I know to check whether the DOM
	// is fully constructed
	else {
	    var currentAnnotationsBatch = {};
    
	    for (var i = 0; i < annotations.length; i++) {
		var current = annotations[i];
		
		// testing purposes
		if (! current.data) {
		    this._debugBadAnnotation(current);
		    continue;
		}
		
		var range = xptrResolveAnnotation(current, window._content.document);
      
		// add to the current batch
		IconManager._addAnnotationToContainer(current, range, null, currentAnnotationsBatch);
      
		// add to the overall collection
		IconManager._addAnnotationToContainer(current, range, null, this.annotations);
	    }

	    // insert this batch
	    this._insertAnnotations(currentAnnotationsBatch);
	}
    }
}

/**
 * Callback function to indicate failure for a loadAnnotations request
 *  - OR - an automatic autoLoad request.  This function indicates failure
 * for one particular annotations server, designated by serverURI.
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri for which annotations were requested
 * @param message string error message
 */
IconManager.prototype.loadAnnotationsOnFailure = function(serverURI, annotatedURI, message) {}

/**
 * Called when an annotation post request succeeds.
 *
 * @param newAnnotation nsIAnnotation the newly created annotation
 */
IconManager.prototype.postAnnotationOnSuccess = function(newAnnotation) {
  try {

    if (this._isInsertIconsEnabled()) {
	var aImg = null;
	
	// try to resolve the annotation
	var aRange = xptrResolveAnnotation(newAnnotation, window._content.document);
	if (aRange) {
	    aImg = this._insertAnnotation(newAnnotation, aRange);
	}
	
	// add an icon and an entry for this
	IconManager._addAnnotationToContainer(newAnnotation, aRange, aImg, this.annotations);
    }
  } catch (e) {
      Components.classes['@mozilla.org/consoleservice;1']
         .getService(Components.interfaces.nsIConsoleService)
            .logStringMessage ("Annozilla: Failure in IconManager.postAnnotationOnSuccess: " + e);
  }
}

/**
 * Called when an annotation post fails.
 *
 * @param message string error message
 */
IconManager.prototype.postAnnotationOnFailure = function(message) {}



/**
 * Called when an annotation update succeeds.
 *
 * @param updatedAnnotation nsIAnnotation that has been updated on server
 */
IconManager.prototype.updateAnnotationOnSuccess = function(updatedAnnotation) {

    if (this._isInsertIconsEnabled()) {
	
	// find the old annotation, if we can
	if (this.annotations[updatedAnnotation.annotationURI]) {
	    var annotObj = this.annotations[updatedAnnotation.annotationURI];
	    
	    // switch to updated annotation
	    annotObj.annot = updatedAnnotation;
	    
	    // update the type, if there is an icon, and it is displaying
	    if (annotObj.img && annotObj.img.parentNode) {
		var type = updatedAnnotation.data.getProperty(IconManager.RDF_NS, "type");
		
		if (type) {
		    type = type.substring(type.indexOf("#") + 1);
		}
		else {
		    type = "Annotation";
		}
		
		annotObj.img.setAttribute("src", IconManager._getIconURI(type));
		annotObj.img.setAttribute("alt", type);
	    }
	}
    }
}

/**
 * Called when an annotation update fails.
 *
 * @param url string url of the annotation that could not be updated
 * @param message string error message
 */
IconManager.prototype.updateAnnotationOnFailure = function(url, message) {}


/**
 * Called when an annotation is successfully deleted.
 *
 * @param url string url of deleted annotation
 */
IconManager.prototype.deleteAnnotationOnSuccess = function(uri) {

    if (this._isInsertIconsEnabled()) {
	// find it 
	if (this.annotations[uri]) {
	    // if there is an img, remove it
	    if (this.annotations[uri].img) {
		IconManager._removeIcon(this.annotations[uri].img);
	    }
	    // delete the object
	    delete this.annotations[uri];
	}
    }
}

/**
 * Called when an annotation delete fails.
 *
 * @param url string url of the annotation that could not be deleted
 * @param message string error message
 */
IconManager.prototype.deleteAnnotationOnFailure = function(url, message) {}

/**
 * Called when the replies to an annotation have been successfully
 * loaded.
 * 
 * @param annotation nsIAnnotation for which replies were requested
 * @param repliesEnum nsISimpleEnumerator containing nsIAnnotation replies
 */
IconManager.prototype.getRepliesOnSuccess = function(annotation, repliesCount, replies) {}

/**
 * Called when getting the replies to an annotation fails.
 *
 * @param annotation nsIAnnotation for which replies were requested
 * @param message string error message
 */
IconManager.prototype.getRepliesOnFailure = function(annotation, message) {}

/**
 * Called an annotation has successfully been replied to.
 * 
 * @param original nsIAnnotation that existed originally
 * @param reply nsIAnnotation reply created
 */
IconManager.prototype.replyToAnnotationOnSuccess = function(original, reply) {}

/**
 * Called when replying to an annotation fails.
 *
 * @param original nsIAnnotation to which we attempted to reply
 * @param message string error message
 */
IconManager.prototype.replyToAnnotationOnFailure = function(original, message) {}

/**
 * Called after a loadAnnotationBody request
 * when the body text of the annotation has been successfully loaded.
 *
 * @param annotation nsIAnnotation with updated body text
 * @param bodyText string text of the annotation's body
 */
IconManager.prototype.loadAnnotationBodyOnSuccess = function(annotation, bodyText) {}

/**
 * Called after a loadAnnotationBody request when
 * the body text fails to load.
 *
 * @param annotation nsIAnnotation that body was requested for (unmodified)
 * @param message string error message
 */
IconManager.prototype.loadAnnotationBodyOnFailure = function(annotation, message) {}

IconManager.prototype.iconClick = function(annotation) {

    var bodyURI;
    
    // if this is a local annotation
    // we must generate a local url for its body
    if (annotation.serverName == this.app.LOCALSERVER_URI) {
	bodyURI = LocalServerUtils.createTempURIFromString(annotation.getBodyText());
    }
    else {
	bodyURI = annotation.bodyURI;
    }

    // TODO This Functionality is duplicated in annozilla.js
    if (!("@mozilla.org/editor/htmleditor;1" in Components.classes) ||
        window.top.document.getElementById ("key_editPage") == undefined) {
        window.openDialog("chrome://annozilla/content/editor/editor.xul",
                          "annotation-body" + Math.random(),
                          "chrome,resizable,dialog=no",
                          annotation,
                          null, null, bodyURI);
    } else {
        window.openDialog("chrome://annozilla/content/edit_body.xul",
                          "annotation-body" + Math.random(),
                          "chrome,resizable,dialog=no",
                          annotation,
                          null, null, bodyURI);
    }
}

/********************
 * PRIVATE FUNCTIONS
 ********************/

IconManager._hasRequiredComponents = function() {
  return true;
}

IconManager.prototype._insertAnnotation = function(annotation, range) {

  // create the icon
  var img = window._content.document.createElementNS(IconManager.XHTML_NS, "img");
  var type = annotation.data.getProperty(IconManager.RDF_NS, "type");
  

  if (type) {
    type = type.substring(type.indexOf("#") + 1);
  }
  else {
	type = "Annotation";
  }

  img.setAttribute("src", IconManager._getIconURI(type));

  // mark the element to be ignored by XPointer resolution
  xptrMarkElement(img);

  IconManager._insertNode(range, img);

  var that = this;

  img.addEventListener("click",
		       function(e) { that.iconClick(annotation);  },
		       false);

  return img;
}

IconManager._getIconURI = function(type) {     
  switch(type) {
  case 'Comment':
    return 'chrome://annozilla/content/icons/annot.png';
    
  case 'Advice':
    return 'chrome://annozilla/content/icons/advice.png';
    
  case 'Example':
    return 'chrome://annozilla/content/icons/example.png';
    
  case 'Explanation':
    return 'chrome://annozilla/content/icons/explanation.png';
    
  case 'Question':
    return 'chrome://annozilla/content/icons/question.png';
    
  case 'SeeAlso':
    return 'chrome://annozilla/content/icons/seealso.png';
    
  default:
    return 'chrome://annozilla/content/icons/annot.png';
  }
}



/**
 * Inserts a node into the Document or DocumentFragment at the start
 * of the Range. If the container is a Text node, this will be split
 * at the start of the Range (as if the Text node's splitText method
 * was performed at the insertion point) and the insertion will occur
 * between the two resulting Text nodes. Adjacent Text nodes will not
 * be automatically merged. If the node to be inserted is a
 * DocumentFragment node, the children will be inserted rather than
 * the DocumentFragment node itself.
 *
 * @param range DOM Range Object
 * @param node DOM Node to insert
 */
IconManager._insertNode = function(range, node_to_insert) {
  var ref_node;   // node beyond insertion point
  var parent;     // parent node
  
  if (range.startContainer.nodeType == node_to_insert.TEXT_NODE || 
      range.startContainer.nodeType == node_to_insert.CDATA_SECTION_NODE) {
    ref_node = range.startContainer.splitText(range.startOffset);
    
    range.startContainer.parentNode.insertBefore(node_to_insert, ref_node);
    return;
  }    
  
  // if the insertion point is at the end of the startContainer node just 
  // append the node to the startContainer node.
  if ( range.startOffset == range.startContainer.childNodes.length) {
    range.startContainer.appendChild( node_to_insert );
  }
  // otherwise, insertion point is not at end of startContainer
  else {
    range.startContainer.insertBefore( node_to_insert,
				       range.startContainer.childNodes.item(range.startOffset));
  }
}

IconManager.prototype._removeAnnotations = function() {
    for (uri in this.annotations) {
	var annotObj = this.annotations[uri];
	if (annotObj.img) {
	    IconManager._removeIcon(annotObj.img);
	}
    }
}

IconManager._removeIcon = function(icon) {
  if (icon.parentNode) {
    icon.parentNode.removeChild(icon);
  }
}

IconManager.prototype._debugBadAnnotation = function(annotation) {

    dump("annotation.data is NULL.  enumerating properties:\n");
    dump("serverName: " + annotation.serverName + "\n");
    dump("annotationURI: " + annotation.annotationURI + "\n");
    dump("bodyURI: " + annotation.bodyURI + "\n");
    dump("getBodyText: " + annotation.getBodyText() + "\n");
    dump("setting bodyText.\n");
    annotation.setBodyText("New test body text");
    dump("new body text: " + annotation.getBodyText() + "\n");
}


IconManager.prototype._insertAnnotations = function(annotations) {
    for (uri in annotations) {
	var annotationObj = annotations[uri];

	var annot = annotationObj.annot;
	if (! annot.data) {
	    // ignore it
	    continue;
	}

	var range = annotationObj.range;
	var img = null;
	if (range) {
	    img = this._insertAnnotation(annot, range);
	}

	this.annotations[uri].img = img;
    }
}

IconManager.prototype._processAnnotationsWithAutoLoadInfo = function(state, annotations) {
   
    if (state == IconManager.PAGE_LOADING) {
	// add the annotations without resolving their context or inserting
      for (var i = 0; i < annotations.length; i++) {
	  IconManager._addAnnotationToContainer(annotations[i], null, null, this.annotations);
      }
    }
    else if (state == IconManager.PAGE_COMPLETED) {
	var currentAnnotationsBatch = {};
	
	for (var i = 0; i < annotations.length; i++) {
	    var current = annotations[i];
	
	    if (! current.data) {
		this._debugBadAnnotation(current);
		continue;
	    }
	    
	    var range = xptrResolveAnnotation(current, window._content.document);
	    
	    // add to the current batch
	    IconManager._addAnnotationToContainer(current, range, null, currentAnnotationsBatch);

	    // add to the overall container
	    IconManager._addAnnotationToContainer(current, range, null, this.annotations);
	}
      
	// insert this batch
	this._insertAnnotations(currentAnnotationsBatch);
    }
    
    // otherwise, the page failed to load, so we ignore annotations
    
}

IconManager._addAnnotationToContainer = function(aAnnotation, aRange, aImg, aContainer) {
  aContainer[aAnnotation.annotationURI] = { annot: aAnnotation,
					    range: aRange,
					    img:   aImg };
}

IconManager.prototype._isInsertIconsEnabled = function() {
    return this.insertIcons;
}

IconManager.prototype._isAutoLoadEnabled = function() {
    return (this.prefs.getPref("autoLoadMulti") != "never");
}

/**
 * Instantiates the Singleton IconManager.
 */
gIconManager = new IconManager();
